/* vi: set sw=2 ts=8: */
/*
 * Copyright 2005 by HI Corporation.
 *
 * This source code may be used for any purpose as long as this
 * copyright message remains in the source code and the documentation
 * for any product made from it includes a note stating:
 *
 * 	"Incorporates software from HI Corporation, copyright 2005."
 * 
 * This code comes WITHOUT warranty of merchantability or fitness for
 * a particular purpose.
 */

/* Sample Program accompanying the Using M3G section of Siggraph 2005
 * course #35 & Eurographics 2006 tutorial T3.
 */

import java.io.IOException;
import java.util.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;
import javax.microedition.midlet.*;
import javax.microedition.m3g.*;

/**
 * The rendering target.
 *
 * Most of the work is done in this class. The m3g file is loaded and
 * displayed and all input is handled here.
 *
 * Use Canvas because GameCanvas swallows key strokes from the
 * number keys.
 */
class TargetCanvas extends Canvas implements Runnable
{
  private World gWorld;
  private Graphics3D g3d;
  boolean hasWorld = false;
  boolean isPlaying = false;

  private Object3D[] loadObjs;
  private CameraManip cameraManip;
  private int w;
  private int h;
  private int fps = 0;
  private int frameCount = 0;
  private long renderTime = 1;
  private String fileName;
    
  public boolean hasException = false;

  // User ids of various nodes
  static final int meshId = 84;
  static final int skinnedMeshId = 86;
  static final int morphingMeshId = 88;
  static final int meshControllerId = 9;
  static final int morphingMeshControllerId = 91;
  static final int skinnedMeshControllerId = 90;

  // Values for keyState
  static final int LEFT_PRESSED = 1 << Canvas.LEFT;
  static final int DOWN_PRESSED = 1 << Canvas.DOWN;
  static final int RIGHT_PRESSED = 1 << Canvas.RIGHT;
  static final int UP_PRESSED = 1 << Canvas.UP;

  // Values for animState
  static final int MESH_ANIM = 0x1;
  static final int MORPHM_ANIM = 0x2;
  static final int SKINM_ANIM = 0x4;

  // Camera movement delta
  static final int DELTA = 10;

  private int animState = MESH_ANIM | MORPHM_ANIM | SKINM_ANIM;
  private int keyState = 0;

  private AnimationController meshController;
  private AnimationController morphingMeshController;
  private AnimationController skinnedMeshController;

  public TargetCanvas(String m3gFile)
  {
    try
    {
      fileName = m3gFile;
      g3d = Graphics3D.getInstance();
      Load();
      w = getWidth();
      h = getHeight();
      cameraManip = new CameraManip(gWorld);
    }
    catch(IOException e)
    {
      System.out.println("loading fails:"+fileName);
      hasException = true;
    }
  }
    

  /* get the file content */
  void Load() throws IOException 
  {
    loadObjs = Loader.load(fileName);
    if (loadObjs==null)
      throw new RuntimeException("M3g file error");
    
    /* find the world node */
    for (int i=0; i<loadObjs.length; ++i) 
    {
      if (loadObjs[i] instanceof World)
      {
	gWorld = (World)loadObjs[i];
	hasWorld = true; 		
	break;
      }	
    } 

    if (!hasWorld)
      throw new RuntimeException("World node not found; incorrect m3g file?");

    meshController = (AnimationController)gWorld.find(meshControllerId);
    morphingMeshController =
	(AnimationController)gWorld.find(morphingMeshControllerId);
    skinnedMeshController =
	(AnimationController)gWorld.find(skinnedMeshControllerId);

    /* Clean up after the loading process. */
    System.gc();
  }
    

  void render()
  {
     if (isPlaying) {
       frameCount++;
       fps = (int)((1000*frameCount) / renderTime) ;
       /* update the scene */
       gWorld.animate((int)renderTime);
     }
     g3d.render(gWorld);
  }
    
  public void startPlay()
  {
    renderTime = 1;
    frameCount = 0;
    isPlaying = true;
  }
    
  public void pausePlay()
  {
    isPlaying = false;
  }


  protected void handleInput()
  {
      int start = 0, end;
      int keyState = getKeyStates();
      int deltaX = 0, deltaY = 0;

      /* Stop & start animation of individual objects by setting
       * active interval on AnimationControllers.
       */
      if (meshController != null) {
	end = (animState & MESH_ANIM) != 0 ? 0 : 1;
	meshController.setActiveInterval(start, end);
      }
      if (skinnedMeshController != null) {
	end = (animState & SKINM_ANIM) != 0 ? 0 : 1;
	skinnedMeshController.setActiveInterval(start, end);
      }
      if (morphingMeshController != null) {
	end = (animState & MORPHM_ANIM) != 0 ? 0 : 1;
	morphingMeshController.setActiveInterval(start, end);
      }
	
      if ((keyState & DOWN_PRESSED) != 0) {
	deltaY -= DELTA;
      }
      if ((keyState & LEFT_PRESSED) != 0) {
	deltaX += DELTA;
      }
      if ((keyState & RIGHT_PRESSED) != 0) {
	deltaX -= DELTA;
      }
      if ((keyState & UP_PRESSED) != 0) {
	deltaY += DELTA;
      }
      if (deltaX != 0 || deltaY != 0)
	cameraManip.rotate( deltaY, deltaX, 0 );
  }


  synchronized protected void paint(Graphics g) 
  {
    if(loadObjs == null) return;
    g.setClip(0, 0, w, h);
    try
    {
      g3d.bindTarget(g);
      g3d.setViewport(0, 0, w, h);
      render();
    } finally { g3d.releaseTarget(); }

    g.setColor(0xffffffff);
    g.drawString("fps: " + fps, 2, 2, g.TOP|g.LEFT);
  }
 
  public void run()
  {
    for(;;) {
      long start, elapsed;
      start = System.currentTimeMillis();
      handleInput();
      repaint();
      elapsed = System.currentTimeMillis() - start;
      // if (want to measure true frame rate)  // Unfriendly to system!! 
      renderTime += (int)elapsed;
      // else {
      //renderTime += (elapsed < 50) ? 50 : (int)elapsed;
      //try {
      //  if (elapsed < 50) Thread.sleep(50-elapsed);
      //} catch (InterruptedException e) { }
      //}
    }
  }
 

  /**
   * Get arrow key state.
   *
   * Returns the current arrow key state, showing which arrow keys have been
   * pressed since the last call. Clears the state.
   */
  int getKeyStates()
  {
    int state = keyState;
    keyState = 0;
    return state;
  }


  /* Receive key pressed messages. */
  protected void keyPressed(int keyCode)
  {
    int action;

    switch (keyCode) {
      case KEY_NUM1:
	animState ^= MESH_ANIM;
	break;
      case KEY_NUM2:
	animState ^= SKINM_ANIM;
	break;
      case KEY_NUM3:
	animState ^= MORPHM_ANIM;
	break;
      default: {
	action = getGameAction(keyCode);
	switch (action) {
      	  case DOWN:
	    keyState |= DOWN_PRESSED;
	    break;
          case LEFT:
	    keyState |= LEFT_PRESSED;
	    break;
          case RIGHT:
	    keyState |= RIGHT_PRESSED;
	    break;
          case UP:
	    keyState |= UP_PRESSED;
	    break;
	  default:
	    break;
	}
      }
    }
  }

  protected void keyReleased(int keyCode)
  {

  }
}

